# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import tempfile
import os
import sys
import unittest

from bs4 import BeautifulSoup
import pytest
from nbval.kernel import CURRENT_ENV_KERNEL_NAME
import jupyter_client.kernelspec as kspec

from nbcelltests.test import run, runWithReturn, runWithReport, runWithHTMLReturn

# Some straightforward TODOs:
#
# - We should generate the notebooks rather than having them as files
#   (same for lint ones). Would also allow for simplification of test
#   class hierarchy.
#
# - Should use assert(Regex)Raises rather than manual try/catch
#
# - Some classes are abstract but not declared as such
#
# - Should split this file up - but only after deciding on
#   organization of modules being tested


# TODO: This test file's manual use of unittest is brittle

CUMULATIVE_RUN = os.path.join(os.path.dirname(__file__), '_cumulative_run.ipynb')
CELL_ERROR = os.path.join(os.path.dirname(__file__), '_cell_error.ipynb')
TEST_ERROR = os.path.join(os.path.dirname(__file__), '_test_error.ipynb')
TEST_FAIL = os.path.join(os.path.dirname(__file__), '_test_fail.ipynb')
COUNTING = os.path.join(os.path.dirname(__file__), '_cell_counting.ipynb')
NONCODE = os.path.join(os.path.dirname(__file__), '_non_code_cell.ipynb')
EMPTY_CELL_WITH_TEST = os.path.join(os.path.dirname(__file__), '_empty_cell_with_test.ipynb')
EMPTYAST_CELL_WITH_TEST = os.path.join(os.path.dirname(__file__), '_emptyast_cell_with_test.ipynb')
SKIPS = os.path.join(os.path.dirname(__file__), '_skips.ipynb')
COVERAGE = os.path.join(os.path.dirname(__file__), '_cell_coverage.ipynb')
CELL_NOT_INJECTED_OR_MOCKED = os.path.join(os.path.dirname(__file__), '_cell_not_injected_or_mocked.ipynb')
BROKEN_MAGICS = os.path.join(os.path.dirname(__file__), '_broken_magics.ipynb')
NO_CODE_CELLS = os.path.join(os.path.dirname(__file__), '_no_code_cells.ipynb')
KERNEL_CWD = os.path.join(os.path.dirname(__file__), '_kernel_cwd.ipynb')

INPUT_CELL_MULTILINE_STRING = os.path.join(os.path.dirname(__file__), '_input_cell_multiline_string.ipynb')
INPUT_TEST_MULTILINE_STRING = os.path.join(os.path.dirname(__file__), '_input_test_multiline_string.ipynb')
INPUT_CELL_NEWLINE_STRING = os.path.join(os.path.dirname(__file__), '_input_cell_newline_string.ipynb')
INPUT_TEST_NEWLINE_STRING = os.path.join(os.path.dirname(__file__), '_input_test_newline_string.ipynb')
INPUT_TEST_INJECTION_COMMENT = os.path.join(os.path.dirname(__file__), '_input_test_injection_comment.ipynb')

# Hack. We want to test expected behavior in distributed situation,
# which we are doing via pytest --forked.
FORKED = '--forked' in sys.argv

# Default to using kernel from current environment (like --current-env of nbval).
TEST_RUN_KW = {
    'current_env': int(os.environ.get("NBCELLTESTS_TESTS_CURRENT_ENV", "1")),
    'kernel_name': os.environ.get("NBCELLTESTS_TESTS_KERNEL_NAME", "")
}


def _assert_undefined(t, name='x'):
    """
    Convenience method to assert that x is not already defined in the kernel.
    """
    t._run("""
    try:
        %s
    except NameError:
        pass
    else:
        raise Exception('%s was already defined')
    """ % (name, name))


def _import_from_path(path, module_name):
    """
    Import and return a python module at the given path, with
    module_name setting __name__.

    See e.g. https://stackoverflow.com/a/67692.
    """
    import importlib.util
    spec = importlib.util.spec_from_file_location(module_name, path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    return mod


def _generate_test_module(notebook, module_name, run_kw=None):
    """
    Generate test file from notebook, then import it with __name__ set
    to module_name, and return it.
    """
    if run_kw is None:
        run_kw = TEST_RUN_KW

    tf = tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False, encoding='utf8')
    tf_name = tf.name
    try:
        # the module name (__name__) doesn't really matter, but
        # will be nbcelltests.tests.test_tests.X, where X is
        # whatever concrete subclass is this method belongs to.
        generated_module = _import_from_path(run(notebook, filename=tf_name, **run_kw), module_name)
    finally:
        try:
            tf.close()
        except Exception:
            pass
        os.remove(tf_name)

    return generated_module


class _TestCellTests(unittest.TestCase):
    # abstract :)

    @classmethod
    def setUpClass(cls):
        assert hasattr(cls, "NBNAME"), "Subclasses must have NBNAME attribute."  # TODO: make actually abstract
        cls.generated_tests = _generate_test_module(notebook=cls.NBNAME, module_name="nbcelltests.tests.%s.%s" % (__name__, cls.__name__))

    def test_coverage(self):
        """
        Subclasses should override this if they want to check coverage.
        """
        assert not hasattr(self, 'test_cell_coverage')


class TestMethodGenerationError(_TestCellTests):
    """Tests of things that should fail during test script generation"""

    @classmethod
    def setUpClass(cls):
        # we're testing what happens in setUpClass
        pass

    def test_non_code_cell_with_test_causes_error(self):
        try:
            _generate_test_module(NONCODE, "module.name.irrelevant")
        except ValueError as e:
            assert e.args[0] == 'Cell 1 is not a code cell, but metadata contains test code!'
        else:
            raise Exception("Test script should fail to generate")

    def test_onlywhitespace_code_cell_with_test_causes_error(self):
        try:
            _generate_test_module(EMPTY_CELL_WITH_TEST, "module.name.irrelevant")
        except ValueError as e:
            assert e.args[0] == 'Code cell 2 is empty, but test contains code.'
        else:
            raise Exception("Test script should fail to generate")

    def test_emptyast_code_cell_with_test_causes_error(self):
        try:
            _generate_test_module(EMPTYAST_CELL_WITH_TEST, "module.name.irrelevant")
        except ValueError as e:
            assert e.args[0] == 'Code cell 2 is empty, but test contains code.'
        else:
            raise Exception("Test script should fail to generate")

    def test_code_not_injected_or_mocked(self):
        try:
            _generate_test_module(CELL_NOT_INJECTED_OR_MOCKED, "module.name.irrelevant")
        except ValueError as e:
            assert e.args[0].startswith('Test 5: cell code not injected into test')
        else:
            raise Exception("Test script should fail to generate")


class TestNoCodeCells(_TestCellTests):
    """Notebook with no code cells."""

    NBNAME = NO_CODE_CELLS

    # the tests are independent, so it's fine to call setUpClass and
    # setUp before every test (and same for tearDown after). When we
    # are generating notebooks, we won't need the single, shared
    # notebook.
    def setUp(self):
        self.t = self.generated_tests.TestNotebook()
        self.t.setUpClass()
        self.t.setUp()

    def tearDown(self):
        self.t.tearDown()
        self.t.tearDownClass()

    def test_no_code_cells(self):
        test_methods = [mthd for mthd in dir(self.t) if mthd.startswith("test_code_cell")]
        assert len(test_methods) == 0


class TestSkips(_TestCellTests):
    """Tests that various conditions result in test methods being generated or not as expected."""

    NBNAME = SKIPS

    # the tests are independent, so it's fine to call setUpClass and
    # setUp before every test (and same for tearDown after). When we
    # are generating notebooks, we won't need the single, shared
    # notebook.
    def setUp(self):
        self.t = self.generated_tests.TestNotebook()
        self.t.setUpClass()
        self.t.setUp()

    def tearDown(self):
        self.t.tearDown()
        self.t.tearDownClass()

    def test_skip_completely_empty_code_cell(self):
        """Empty cell and test"""
        assert not hasattr(self.t, "test_code_cell_1")

    def test_skip_no_code_code_cell(self):
        """Cell's just a comment, and no test."""
        assert not hasattr(self.t, "test_code_cell_2")

    def test_noskip_with_no_test_field_in_metadata(self):
        """Test where cell has no test metadata is just cell"""
        self.t.test_code_cell_3()
        self.t._run("assert x == 3")

    def test_noskip_with_completely_empty_test(self):
        """Test is just %cell where test is completely empty"""
        self.t.test_code_cell_4()
        self.t._run("assert x == 4")

    def test_noskip_with_deliberate_no_cell_injection(self):
        """Runs test but not cell"""
        self.t.test_code_cell_5()
        self.t._run("assert x == 6")  # value from test, not cell


class TestCumulativeRun(_TestCellTests):
    """Tests that state in notebook is built up correctly."""

    NBNAME = CUMULATIVE_RUN

    def test_state(self):
        """In the expected case, the five cells of the notebook should work
        out as follows:

             cell    test   state
        1:    -       -       -
        2:   x=0      -      x=0
        3:   x+=1     -      x=1
        4:   x+=1     -      x=2
        5:    -      x+=1    x=3

        However, these tests originally detected repeated execution of
        cells in the same kernel. Cell 2 actually only sets x to 0 if
        x is not already defined; if x is already defined, cell 2 does
        not change x's value. If for some reason each test results in
        multiple executions of the cells, the results will look like
        this:

             cell    test   state
        1:    -       -       -
        2:    -       -      x>0   (bad; x was previously defined in kernel)
        3:   x+=1     -      x>1   (bad)
        4:   x+=1     -      x>2   (bad)
        5:    -      x+=1    x>3   (bad)
        """
        # this is one long method rather individual ones because of
        # building up state

        t = self.generated_tests.TestNotebook()
        t.setUpClass()

        # check cell did not run
        # (deliberate no cell in test, no code in test)
        # TODO: should this be no method, as there's only empty ast in the test
        t.setUp()
        _assert_undefined(t, 'x')
        t.test_code_cell_1()
        _assert_undefined(t, 'x')
        t.tearDown()
        if FORKED:
            t.tearDownClass()

        # check cell ran
        # (%cell in test)
        if FORKED:
            t.setUpClass()
        t.setUp()
        t.test_code_cell_2()
        t._run("""
        assert x == 0, x
        """)
        t.tearDown()
        if FORKED:
            t.tearDownClass()

        # check cumulative cells ran
        if FORKED:
            t.setUpClass()
        t.setUp()
        t.test_code_cell_3()
        t._run("""
        assert x == 1, x
        """)
        t.tearDown()
        if FORKED:
            t.tearDownClass()

        # check cumulative cells ran (but not multiple times!)
        if FORKED:
            t.setUpClass()
        t.setUp()
        t.test_code_cell_4()
        t._run("""
        assert x == 2, x
        """)
        t.tearDown()
        if FORKED:
            t.tearDownClass()

        # check test affects state
        if FORKED:
            t.setUpClass()
        t.setUp()
        t.test_code_cell_5()
        t._run("""
        assert x == 3, x
        """)
        t.tearDown()
        if FORKED:
            t.tearDownClass()

        # test defaults to %cell
        if FORKED:
            t.setUpClass()
        t.setUp()
        t.test_code_cell_6()
        t._run("""
        assert y == 10
        """)
        t.tearDown()
        if FORKED:
            t.tearDownClass()

        # deliberate no %cell; cell 8 will check it's also not subsequently run
        # i.e. a will never be defined
        if FORKED:
            t.setUpClass()
        t.setUp()
        t.test_code_cell_7()
        _assert_undefined(t, 'a')
        t.tearDown()
        if FORKED:
            t.tearDownClass()

        # check cell 7 above did not run
        if FORKED:
            t.setUpClass()
        t.setUp()
        t.test_code_cell_8()
        t._run("""
        assert z == 1
        assert y == 10
        """)
        _assert_undefined(t, 'a')
        t.tearDown()

        t.tearDownClass()


class TestExceptionInCell(_TestCellTests):
    """Tests related to exceptions in cells"""

    NBNAME = CELL_ERROR

    def test_exception_in_cell_is_detected(self):
        t = self.generated_tests.TestNotebook()

        t.setUpClass()
        t.setUp()

        # cell should error out
        try:
            t.test_code_cell_1()
        except Exception as e:
            assert e.args[0].startswith("Running cell+test for code cell 1; execution caused an exception")
            assert e.args[0].endswith("My code does not even run")
        else:
            raise Exception("Cell should have errored out")
        finally:
            t.tearDown()
            t.tearDownClass()


class TestExceptionInTest(_TestCellTests):
    """Tests related to exeptions in tests"""

    NBNAME = TEST_ERROR

    def test_exception_in_test_is_detected(self):
        t = self.generated_tests.TestNotebook()
        t.setUpClass()
        t.setUp()

        # caught cell error
        t.test_code_cell_1()

        t.tearDown()
        if FORKED:
            t.tearDownClass()

        if FORKED:
            t.setUpClass()
        t.setUp()

        _assert_undefined(t, 'x')

        # test should error out
        try:
            t.test_code_cell_2()
        except Exception as e:
            assert e.args[0].startswith("Running cell+test for code cell 2; execution caused an exception")
            assert e.args[0].endswith("My test is bad too")
        else:
            raise Exception("Test should have failed")
        finally:
            t.tearDown()
            t.tearDownClass()


class TestFailureInTest(_TestCellTests):
    """Tests related to test failures"""

    NBNAME = TEST_FAIL

    def test_failure_is_detected(self):
        """Test expected to fail - make sure we detect that."""
        t = self.generated_tests.TestNotebook()
        t.setUpClass()
        t.setUp()

        try:
            t.test_code_cell_1()
        except Exception as e:
            assert e.args[0].startswith("Running cell+test for code cell 1; execution caused an exception")
            assert e.args[0].endswith("x should have been -1 but was 1")
        else:
            raise Exception("Test should have failed")

        t.tearDown()
        if FORKED:
            t.tearDownClass()

        if FORKED:
            t.setUpClass()
        t.setUp()
        # subsequent cell should also fail
        try:
            t.test_code_cell_2()
        except Exception as e:
            assert e.args[0].startswith("Running cell+test for code cell 1; execution caused an exception")
            assert e.args[0].endswith("x should have been -1 but was 1")
        else:
            raise Exception("Test should have failed at cell 1")

        t.tearDown()
        t.tearDownClass()


# TODO: see https://github.com/computationalmodelling/nbval/issues/147
class TestSomeSanity(_TestCellTests):

    NBNAME = BROKEN_MAGICS

    @pytest.mark.xfail(reason="error messages on shell channel are ignored by nbval")
    def test_bad_magic_does_let_everything_pass(self):
        t = self.generated_tests.TestNotebook()

        t.setUpClass()
        t.setUp()

        # cell should error out
        try:
            t.test_code_cell_1()
        except Exception as e:
            assert e.args[0].startswith("UsageError: Line magic function `%magics2` not found.")
        else:
            raise Exception("Cell should have errored out")
        finally:
            t.tearDown()
            t.tearDownClass()


class TestCellCounting(_TestCellTests):
    """Check that various things don't throw off cell+test correspondence."""

    NBNAME = COUNTING

    # the tests are independent, so it's fine to call setUpClass and
    # setUp before every test (and same for tearDown after). When we
    # are generating notebooks, we won't need the single, shared
    # notebook.
    def setUp(self):
        self.t = self.generated_tests.TestNotebook()
        self.t.setUpClass()
        self.t.setUp()

    def tearDown(self):
        self.t.tearDown()
        self.t.tearDownClass()

    def test_skips(self):
        """
        There's an empty cell at the start of the notebook to make sure skips don't
        affect test/cell correspondence.
        """
        assert not hasattr(self.t, "test_code_cell_1")

    def test_still_ok_after_markdown(self):
        """Correspondence still ok after markdown cell?"""
        self.t.test_code_cell_2()

    def test_still_ok_after_raw(self):
        """Correspondence still ok after raw cell?"""
        self.t.test_code_cell_3()

    def test_count(self):
        """No unexpected extra test methods"""
        test_methods = [mthd for mthd in dir(self.t) if mthd.startswith("test_code_cell")]
        self.assertListEqual(sorted(test_methods), ['test_code_cell_2', 'test_code_cell_3'])


class TestCellCoverage(_TestCellTests):
    """Test that test_coverage works."""

    NBNAME = COVERAGE

    # the tests are independent, so it's fine to call setUpClass and
    # setUp before every test (and same for tearDown after). When we
    # are generating notebooks, we won't need the single, shared
    # notebook.
    def setUp(self):
        self.t = self.generated_tests.TestNotebook()
        self.t.setUpClass()
        self.t.setUp()

    def tearDown(self):
        self.t.tearDown()
        self.t.tearDownClass()

    def test_coverage(self):
        """
        There's a skipped test at the start of the notebook to make sure skips don't
        affect test/cell correspondence.
        """
        try:
            self.t.test_cell_coverage()
        except AssertionError as e:
            assert e.args[0] == 'Actual cell coverage 25.0 < minimum required of 50'
        else:
            raise ValueError("Cell coverage test should have failed.")


class _TestInput(unittest.TestCase):
    """Various input that breaks celltests."""

    # abstract

    @classmethod
    def setUpClass(cls):
        assert hasattr(cls, "NBNAME"), "Subclasses must have NBNAME attribute."  # TODO: make actually abstract
        cls.generated_tests = _generate_test_module(notebook=cls.NBNAME, module_name="nbcelltests.tests.%s.%s" % (__name__, cls.__name__))

    def setUp(self):
        self.t = self.generated_tests.TestNotebook()
        self.t.setUpClass()
        self.t.setUp()

    def tearDown(self):
        self.t.tearDown()
        self.t.tearDownClass()


class TestInputCellMultilineString(_TestInput):

    NBNAME = INPUT_CELL_MULTILINE_STRING

    def test_input_cell_multiline_string(self):
        self.t.test_code_cell_1()
        expected = "\nanything\n"
        self.t._run('assert problemc == """%s"""' % expected)


class TestInputTestMultilineString(_TestInput):

    NBNAME = INPUT_TEST_MULTILINE_STRING

    def test_input_celltest_multiline_string(self):
        self.t.test_code_cell_1()
        self.t._run("assert x == 1")
        expected = "\nanything\n"
        self.t._run('assert problemt == """%s"""' % expected)


class TestInputCellNewlineString(_TestInput):

    NBNAME = INPUT_CELL_NEWLINE_STRING

    def test_input_cell_newline_string(self):
        self.t.test_code_cell_1()
        expected = "\\n"
        self.t._run('assert problem == "%s"' % expected)


class TestInputTestNewlineString(_TestInput):

    NBNAME = INPUT_TEST_NEWLINE_STRING

    def test_input_celltest_newline_string(self):
        self.t.test_code_cell_1()
        self.t._run("assert problemc == 'x'")
        expected = "\\n"
        self.t._run("assert problemt == '%s'" % expected)


class TestInputTestInjectionComment(_TestInput):

    NBNAME = INPUT_TEST_INJECTION_COMMENT

    def test_input_celltest_injection_comment(self):
        self.t.test_code_cell_1()
        # though it's in the nb test itself, make sure the test
        # actually ran
        self.t._run("assert x == 1")


class TestKernelCwd(_TestCellTests):
    """Do we remember to set cwd?"""

    NBNAME = KERNEL_CWD

    # the tests are independent, so it's fine to call setUpClass and
    # setUp before every test (and same for tearDown after). When we
    # are generating notebooks, we won't need the single, shared
    # notebook.
    def setUp(self):
        self.t = self.generated_tests.TestNotebook()
        self.t.setUpClass()
        self.t.setUp()

    def tearDown(self):
        self.t.tearDown()
        self.t.tearDownClass()

    def test_file_in_nb_dir(self):
        """file in same dir as nb, which nb will look for"""
        target_file = os.path.join(os.path.dirname(self.NBNAME), "hello.txt")
        if os.path.exists(target_file):
            raise ValueError("Going to generate %s but it already exists." % target_file)

        try:
            open(target_file, 'w').close()
            self.t.test_code_cell_1()
        finally:
            try:
                os.remove(target_file)
            except BaseException:
                pass


# some cryptic interface going on here, could be improved :)

@pytest.mark.parametrize(
    "notebook, current_env, kernel_name, exception, expected_text", [
        # the notebook's own kernel
        ('_kernel_check.ipynb', False, "", kspec.NoSuchKernel, "NOBODY WOULD EVER CALL A KERNEL THIS"),
        # current env's kernel
        ('_kernel_check.ipynb', True, "", None, CURRENT_ENV_KERNEL_NAME),
        # named kernel
        ('_kernel_check.ipynb', False, "OR THIS", kspec.NoSuchKernel, None),
        # default kernel if none specified or in nb
        ('_kernel_check1.ipynb', False, "python3", None, None),
        # inconsistent request
        ('_kernel_check.ipynb', True, "something", ValueError, "mutually exclusive"),
    ]
)
def test_kernel_selection(notebook, current_env, kernel_name, exception, expected_text):
    if not expected_text:
        expected_text = kernel_name

    nb = os.path.join(os.path.dirname(__file__), notebook)

    def make_test_mod():
        return _generate_test_module(nb, module_name="nbcelltests.tests.%s.%s" % (__name__, "test_kernel_selection"),
                                     run_kw=dict(current_env=current_env, kernel_name=kernel_name))

    if exception:
        try:
            test_mod = make_test_mod()
            test_mod.TestNotebook.setUpClass()
        except exception as e:
            assert e.args[0].endswith(expected_text)
        else:
            raise ValueError("Expected exception %s(%s) to be raised", (exception, expected_text))
    else:
        test_mod = make_test_mod()
        test_mod.TestNotebook.setUpClass()
        assert test_mod.TestNotebook.kernel.km.kernel_name == expected_text

######


# TODO: there's a repeated pattern of cleaning up files generated
# during tests, but address
# https://github.com/jpmorganchase/nbcelltests/issues/125 before
# deciding how to handle that better here.

# runWithReturn

def test_basic_runWithReturn_pass():
    """Basic check - just that it runs without error"""
    generates = os.path.join(os.path.dirname(__file__), "_cell_coverage_test.py")
    if os.path.exists(generates):
        raise ValueError("Going to generate %s but it already exists." % generates)

    try:
        _ = runWithReturn(COVERAGE, rules={'cell_coverage': 10}, **TEST_RUN_KW)
    finally:
        try:
            os.remove(generates)
        except Exception:
            pass


def test_basic_runWithReturn_fail():
    """Basic check - just that it fails"""
    generates = os.path.join(os.path.dirname(__file__), "_cell_coverage_test.py")
    if os.path.exists(generates):
        raise ValueError("Going to generate %s but it already exists." % generates)

    try:
        _ = runWithReturn(COVERAGE, rules={'cell_coverage': 100}, **TEST_RUN_KW)
    except Exception:
        pass  # would need to alter run fn or capture output to check more exactly
    else:
        raise ValueError("coverage check should have failed, but didn't")
    finally:
        try:
            os.remove(generates)
        except Exception:
            pass


# runWithReport

def test_basic_runWithReport_pass():
    """Basic check - just that it runs without error"""
    generates = os.path.join(os.path.dirname(__file__), "_cell_coverage_test.py")
    if os.path.exists(generates):
        raise ValueError("Going to generate %s but it already exists." % generates)

    from nbcelltests.define import TestType
    try:
        ret = runWithReport(COVERAGE, executable=None, rules={'cell_coverage': 10}, **TEST_RUN_KW)
    finally:
        try:
            os.remove(generates)
        except Exception:
            pass

    assert len(ret) == 1
    assert (ret[0].passed, ret[0].type, ret[0].message) == (1, TestType.CELL_COVERAGE, 'Testing cell coverage')


# def test_basic_runWithReport_fail():
#    from nbcelltests.define import TestType
#    # TODO it fails here, but it shouldn't, right? we want to be able to report
#    ret = runWithReport(COVERAGE, executable=None, rules={'cell_coverage':100})
#    assert len(ret) == 1
#    assert (ret[0].passed, ret[0].type, ret[0].message) == (False, TestType.CELL_COVERAGE, 'Testing cell coverage')


# runWithHTMLReturn

def test_basic_runWithHTMLReturn_pass():
    """Check it runs without error and generates the expected files and html."""
    generates = [os.path.join(os.path.dirname(__file__), "_cell_coverage_test.py"),
                 os.path.join(os.path.dirname(__file__), "_cell_coverage_test.html")]
    exists_check = [os.path.exists(f) for f in generates]
    if any(exists_check):
        raise ValueError("Going to generate %s but already exist(s)" % [f for f, exists in zip(generates, exists_check) if exists])

    try:
        ret = runWithHTMLReturn(COVERAGE, executable=None, rules={'cell_coverage': 10}, **TEST_RUN_KW)

        for f in generates:
            assert os.path.exists(f), "Should have generated %s but did not" % f
    finally:
        try:
            for f in generates:
                os.remove(f)
        except Exception:
            pass

    _check(ret, coverage_result="Passed")


def test_basic_runWithHTMLReturn_fail():
    """Check it runs without error and generates the expected files and html."""
    generates = [os.path.join(os.path.dirname(__file__), "_cell_coverage_test.py"),
                 os.path.join(os.path.dirname(__file__), "_cell_coverage_test.html")]
    exists_check = [os.path.exists(f) for f in generates]
    if any(exists_check):
        raise ValueError("Going to generate %s but already exist(s)" % [f for f, exists in zip(generates, exists_check) if exists])

    try:
        ret = runWithHTMLReturn(COVERAGE, executable=None, rules={'cell_coverage': 100}, **TEST_RUN_KW)

        for f in generates:
            assert os.path.exists(f), "Should have generated %s but did not" % f
    finally:
        try:
            for f in generates:
                os.remove(f)
        except Exception:
            pass

    _check(ret, coverage_result="Failed")


def _check(html, coverage_result):
    """
    Check html report contains expected results.

    coverage_result is 'Passed' or 'Failed' depending what you want to
    check.
    """
    html_soup = BeautifulSoup(html, "html.parser")

    tests_ran = False
    for p in html_soup.find_all("p"):
        # 1 cell test plus coverage test
        if p.text.startswith("5 tests ran in"):
            tests_ran = True
            break

    assert tests_ran

    expected_results = {
        "test_code_cell_2": "Passed",
        "test_code_cell_3": "Passed",
        "test_code_cell_4": "Passed",
        "test_code_cell_5": "Passed",
        "test_cell_coverage": coverage_result,
    }

    actual_results = html_soup.find_all(class_="results-table-row")

    assert len(actual_results) == len(expected_results)

    for actual_result, expected_name in zip(sorted(actual_results, key=lambda x: x.find_next(class_="col-name").text),
                                            sorted(expected_results)):
        name = actual_result.find_next(class_="col-name").text
        state = actual_result.find_next(class_="col-result").text

        assert name.endswith(expected_name)
        assert state == expected_results[expected_name]


# TODO: either make genuinely abstract, or don't use classes/inheritance at all here (since classes/inheritance are not meaningful here anyway).
del _TestCellTests
del _TestInput



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# nbcelltests documentation build configuration file, created by
# sphinx-quickstart on Fri Jan 12 22:07:11 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import sys
import os
import os.path
import subprocess
import sphinx_rtd_theme
from recommonmark.transform import AutoStructify
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.coverage', 'sphinx.ext.viewcode', 'sphinx.ext.autodoc', 'sphinx.ext.napoleon', 'recommonmark']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ['.rst', '.md']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'nbcelltests'
copyright = '2018, Tim Paine'
author = 'Tim Paine'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

_version_py = os.path.join('..', 'nbcelltests', '_version.py')
version_ns = {}

with open(_version_py, mode='r') as version_file:
    exec(version_file.read(), version_ns)

# The short X.Y version.
version = '%i.%i' % version_ns['version_info'][:2]
# The full version, including alpha/beta/rc tags.
release = version_ns['__version__']


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'nbcelltestsdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'nbcelltests.tex', 'nbcelltests Documentation',
     'Tim Paine', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'nbcelltests', 'nbcelltests Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'nbcelltests', 'nbcelltests Documentation',
     author, 'nbcelltests', 'One line description of project.',
     'Miscellaneous'),
]

def run_copyreadme(_):
    out = os.path.abspath(os.path.join(os.path.dirname(__file__), 'index.md'))
    readme = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'README.md'))
    api = os.path.abspath(os.path.join(os.path.dirname(__file__), 'api.md'))
    with open(out, 'w') as fp1:
        with open(readme, 'r') as fp2:
            for line in fp2:
                if 'src=' in line:
                    # <img>
                    fp1.write(line.replace("docs/", ""))
                elif "](docs/" in line:
                    # md
                    fp1.write(line.replace("](docs/", "]("))
                else:
                    fp1.write(line)

        fp1.write("# API Documentation\n\n")
        with open(api, 'r') as fp2:
            fp1.write(fp2.read())


def run_apidoc(_):
    out_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'api'))
    nbcelltests_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'nbcelltests'))
    cmd_path = 'sphinx-apidoc'
    if hasattr(sys, 'real_prefix'):  # Check to see if we are in a virtualenv
        # If we are, assemble the path manually
        cmd_path = os.path.abspath(os.path.join(sys.prefix, 'bin', 'sphinx-apidoc'))
    subprocess.check_call([cmd_path,
                           '-E',
                           '-M',
                           '-o',
                           out_dir,
                           nbcelltests_dir,
                           '--force'])


def setup(app):
    app.add_config_value('recommonmark_config', {
        'auto_toc_tree_section': 'Contents',
    }, True)
    app.add_transform(AutoStructify)
    app.connect('builder-inited', run_copyreadme)
    app.connect('builder-inited', run_apidoc)



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

# TODO go back and get version of nbval it was copied from, and record
# what modifications have been made.

# This file includes code copied from nbval under the following license:
# Copyright (C) 2014  Oliver W. Laslett  <O.Laslett@soton.ac.uk>
#               David Cortes-Ortuno
#           Maximilian Albert
#           Ondrej Hovorka
#           Hans Fangohr
#           (University of Southampton, UK)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# Neither the names of the contributors nor the associated institutions
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

try:
    from Queue import Empty
except ImportError:
    from queue import Empty
import logging
import unittest
import os

import nbformat
from nbval.kernel import RunningKernel, CURRENT_ENV_KERNEL_NAME

from nbcelltests.shared import empty_ast, cell_injected_into_test, source2lines, lines2source, get_cell_inj_span, get_test, only_whitespace, CELL_SKIP_TOKEN, CELL_INJ_TOKEN


def _inject_cell_into_test(cell_source, test_source):
    """Inserts cell_source into test.

    The cell_source is inserted wherever %cell appears at the stripped
    start of a line.

    All lines of cell_source are prefixed with the same leading
    whitespace as %cell.

    Example:

      test:
        if x > 0:
            %cell # end of %cell
        assert x == 10
        assert z == 15

      cell:
        x*=5
        z = x + \
            5
        print(x)

      celltest:
        if x > 0:
            x*=5
            z = x + \
                5
            print(x) # end of %cell
        assert x == 10
        assert z == 15
    """
    celltest_lines = []
    for test_line in source2lines(test_source):
        cell_inj_span = get_cell_inj_span(test_line)
        if cell_inj_span is not None:
            prefix = test_line[0:cell_inj_span[0]]
            for cell_line in source2lines(cell_source):
                celltest_lines.append(prefix + cell_line)

            suffix = test_line[cell_inj_span[1]::]
            if len(suffix) > 0:
                if len(celltest_lines) == 0:
                    celltest_lines.append('')
                celltest_lines[-1] += suffix
        else:
            celltest_lines.append(test_line)
    return lines2source(celltest_lines)


def get_celltests(path_to_notebook):
    """
    Return a dictionary of {code cell number: celltest_info} for all
    non-empty code cells in the given notebook.

    celltest_info is a dictionary containing:

      * 'source' string of the test supplied for a cell, plus the
        cell itself wherever injected into the test using %cell.

      * 'cell_injected' flag indicating whether the cell was injected
        into the test
    """
    notebook = nbformat.read(path_to_notebook, 4)
    celltests = {}
    code_cell = 0
    for i, cell in enumerate(notebook.cells, start=1):
        test_source = lines2source(get_test(cell))
        test_ast_empty = empty_ast(_inject_cell_into_test("pass", test_source))

        if cell.get('cell_type') != 'code':
            if not test_ast_empty:
                raise ValueError("Cell %d is not a code cell, but metadata contains test code!" % i)
            continue

        code_cell += 1

        if empty_ast(cell['source']):  # TODO: maybe this should be only_whitespace?
            if not test_ast_empty:
                raise ValueError("Code cell %d is empty, but test contains code." % code_cell)
            continue

        cell_injected = cell_injected_into_test(test_source)

        if only_whitespace(test_source):
            celltest = cell['source']
            cell_injected = True
        elif cell_injected is None:
            raise ValueError(
                r"Test {}: cell code not injected into test; either add '{}' to the test, or add '{}' to deliberately suppress cell execution".format(
                    code_cell, CELL_INJ_TOKEN, CELL_SKIP_TOKEN))
        elif cell_injected is False:
            celltest = test_source
        else:
            celltest = _inject_cell_into_test(cell['source'], test_source)

        celltests[code_cell] = {'source': celltest, 'cell_injected': cell_injected}

    return celltests


def generate_name(testcase_func, param_num, param):
    """Used to generate parameterized method names like test_code_cell_n"""
    return "test_code_cell_%s" % param.args[0]


class TestNotebookBase(unittest.TestCase):
    """Base class for representing a notebook's code cells and their
    associated tests; can submit cells to a kernel, track which cells
    have been executed, and ensure all necessary cells are executed in
    order.

    For instance, requesting to run test_code_cell_7,
    test_code_cell_8, and test_code_cell_9 (in that order) will result
    in:

      1. test_code_cell_7: executes cells 1, 2, 3, 4, 5, 6, 7
      2. test_code_cell_8: executes cell 8
      3. test_code_cell_9: executes cell 9

    The above assumes all cells+tests succeed.

    If e.g. cell 3 fails, the above changes to:

      1. test_code_cell_7: executes cells 1, 2; fails on 3 (with a
                           message that cell 3 failed)
      2. test_code_cell_8: also fails on 3
      3. test_code_cell_9: also fails on 3

    Requesting to run test_code_cell_5 and test_code_cell_3 (in that order)
    will result in:

      1. test_code_cell_5: executes cells 1, 2, 3, 4, 5 (test passes)
      2. test_code_cell_3: execute nothing (test passes)


    When does a cell execute?
    -------------------------

    A cell's test will execute its corresponding cell wherever %cell
    appears at the start of a stripped line.

    If a cell does not have a test defined, or the test is only
    whitespace, the cell will be executed (i.e. the test will be
    assumed to be just %cell).

    If a cell's test contains a "# no %cell" comment line, the cell
    will not be executed (though the test will). Use for mocking out a
    whole cell (ideally it's better to avoid this and mock things the
    cell uses, but that might not be possible).

    Code cells that have an empty ast are considered empty. An
    exception will be raised if attempting to test empty code cells or
    non-code cells.


    Notes
    -----

    This is an abstract class; subclasses will supply the source of
    code cells and their associated tests in celltests, plus test
    methods as entry points for test runners.

    'cell' used in this class refers to cell number; 'cell content'
    typically refers to code_cell+test (depending what is passed in).

    """
    # abstract - subclasses will define KERNEL_NAME and celltests
    # (TODO: make actually abstract...)

    @classmethod
    def setUpClass(cls):
        cls.celltests_run = set()
        # the rest is like nbval's IPyNbFile.setup() (or will be...)
        if cls._current_env and cls._kernel_name:
            raise ValueError("current_env and kernel_name are mutually exclusive")
        if cls._current_env:
            kernel_name = CURRENT_ENV_KERNEL_NAME
        elif cls._kernel_name:
            kernel_name = cls._kernel_name
        else:
            notebook = nbformat.read(cls._notebook, 4)
            kernel_name = notebook['metadata'].get('kernelspec', {}).get('name', 'python')
        cls.kernel = RunningKernel(kernel_name, os.path.dirname(cls._notebook))

    @classmethod
    def tearDownClass(cls):
        cls.kernel.stop()

    def assert_coverage(self, cells_covered, min_required):
        assert cells_covered >= min_required, "Actual cell coverage %s < minimum required of %s" % (cells_covered, min_required)

    def run_test(self, cell):
        """
        Run any cells preceding cell (number) that have not already been
        run, then run cell itself.
        """
        preceding_cells = set(range(1, cell)) & self.celltests.keys()
        for preceding_cell in sorted(set(preceding_cells) - self.celltests_run):
            self._run_cell(preceding_cell)
        self._run_cell(cell)
        if not self.celltests[cell]['cell_injected']:
            # TODO: this will appear in the html report under the test
            # method as captured logging, but it would be better
            # reported as a warning by pytest. However, (a) pytest is
            # already reporting various spurious warnings
            # (e.g. traitlets deprecations), and (b) pytest warnings
            # are not being reported in the html we show right now.
            logging.warning("Cell %d was not executed as part of the cell test", cell)

    def _run_cell(self, cell):
        """Run cell and record its execution"""
        self._run(self.celltests[cell]["source"], "Running cell+test for code cell %d" % cell)
        self.celltests_run.add(cell)

    def _run(self, cell_content, description=''):
        """
        Send supplied cell_content (cell source string) to kernel and
        check it runs without exception.
        """
        # Start of code from nbval (with modifications)
        # https://github.com/computationalmodelling/nbval
        #
        # Modifications:
        #   * ? (things before 2020)
        #   * Add description to exception messages, so it's easy to see which
        #     cell is failing.
        msg_id = self.kernel.execute_cell_input(cell_content, allow_stdin=False)

        # Poll the shell channel to get a message
        try:
            self.kernel.await_reply(msg_id)
        except Empty:
            raise Exception('%s; Kernel timed out waiting for message!' % description)

        while True:
            # The iopub channel broadcasts a range of messages. We keep reading
            # them until we find the message containing the side-effects of our
            # code execution.
            try:
                # Get a message from the kernel iopub channel
                msg = self.kernel.get_message(stream='iopub')

            except Empty:
                raise Exception('%s; Kernel timed out waiting for message!' % description)

            # now we must handle the message by checking the type and reply
            # info and we store the output of the cell in a notebook node object
            msg_type = msg['msg_type']
            reply = msg['content']

            # Is the iopub message related to this cell execution?
            if msg['parent_header'].get('msg_id') != msg_id:
                continue

            # When the kernel starts to execute code, it will enter the 'busy'
            # state and when it finishes, it will enter the 'idle' state.
            # The kernel will publish state 'starting' exactly
            # once at process startup.
            if msg_type == 'status':
                if reply['execution_state'] == 'idle':
                    break
                else:
                    continue
            elif msg_type == 'execute_input':
                continue
            elif msg_type.startswith('comm'):
                continue
            elif msg_type == 'execute_reply':
                continue
            elif msg_type in ('display_data', 'execute_result'):
                continue
            elif msg_type == 'stream':
                continue

            # if the message type is an error then an error has occurred during
            # cell execution. Therefore raise a cell error and pass the
            # traceback information.
            elif msg_type == 'error':
                traceback = '\\n' + '\\n'.join(reply['traceback'])
                msg = "%s; execution caused an exception" % description
                raise Exception(msg + '\\n' + traceback)

            # any other message type is not expected
            # should this raise an error?
            else:
                print("%s; unhandled iopub msg:" % description, msg_type)

        # End of code from nbval


# Fetches notebook source at import time. Don't necessarily think we
# should do the dynamic test generation this way; first priority was
# just to clean up so code's not built up in string.

BASE = '''
from parameterized import parameterized
from nbcelltests.tests_vendored import TestNotebookBase, get_celltests, generate_name

_notebook = r"{path_to_notebook}"
_celltests = get_celltests(_notebook)

class TestNotebook(TestNotebookBase):

    _current_env = {current_env}
    _kernel_name = "{kernel_name}"
    _notebook = _notebook
    celltests = _celltests

    @parameterized.expand([(i,) for i in _celltests], name_func=generate_name, skip_on_empty=True)
    def _test_code_cell(self, cell_num):
        self.run_test(cell_num)

    @parameterized.expand({coverage}, skip_on_empty=True, name_func=lambda *args: "test_cell_coverage")
    def _test_coverage(self, actual,required):
        self.assert_coverage(actual,required)
'''

JSON_CONFD = '''
import pytest
import json
import io

class JsonReporter:
    def __init__(self, config):
        self.config = config
        self.verbosity = self.config.option.verbose
        self.serializable_collection_reports = []
        self.serializable_test_reports = []
        self.serializable_collect_items = []

    # ---- These functions store captured test items and reports ----

    def pytest_runtest_logreport(self, report):
        """Store all test reports for evaluation on finish"""
        data = self.config.hook.pytest_report_to_serializable(
            config=self.config, report=report
        )
        self.serializable_test_reports.append(data)

    def pytest_collectreport(self, report):
        """Store all collected reports for evaluation on finish
        """
        data = self.config.hook.pytest_report_to_serializable(
            config=self.config, report=report
        )
        self.serializable_collection_reports.append(data)

    def pytest_collection_finish(self, session):
        if self.config.getoption("collectonly"):
            self.serializable_collect_items = [
                dict(nodeid=i.nodeid)
                for i in session.items
            ]


    # ---- Code below writes up report ----

    @pytest.hookimpl(hookwrapper=True)
    def pytest_sessionfinish(self, exitstatus):
        """Called when test session has finished.
        """
        yield
        with io.open(self.config.getoption("jsonpath"), "w", encoding="utf-8") as fp:
            json.dump(
                dict(
                    reports=self.serializable_collection_reports +
                        self.serializable_test_reports,
                    collected_items=self.serializable_collect_items,
                ),
                fp)

def pytest_configure(config):
    reporter = JsonReporter(config)
    config.pluginmanager.register(reporter, 'jsonreporter')

def pytest_addoption(parser):
    term_group = parser.getgroup("terminal reporting")
    term_group._addoption(
        '--internal-json-report', action='store', dest='jsonpath',
        metavar='path', default=None,
        help='create JSON report file at given path.')

'''



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
# for Coverage
from nbcelltests import _jupyter_server_extension_paths


class TestInit:
    def test__jupyter_server_extension_paths(self):
        assert _jupyter_server_extension_paths() == [{"module": "nbcelltests"}]



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
# for Coverage
from mock import MagicMock
from nbcelltests import load_jupyter_server_extension


class TestExtension:
    def test_load_jupyter_server_extension(self):

        m = MagicMock()

        m.web_app.settings = {}
        m.web_app.settings['base_url'] = '/test'
        load_jupyter_server_extension(m)



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
from collections import namedtuple
import os
from operator import itemgetter
from bs4 import BeautifulSoup
import pytest

from nbcelltests.lint import lint_lines_per_cell, lint_cells_per_notebook, lint_function_definitions, lint_class_definitions, lint_kernelspec, lint_magics, run, runWithHTMLReturn
from nbcelltests.define import LintType

LR = namedtuple("lint_result", ['passed', 'type'])

# note that list comparison with = in this file assumes pytest


@pytest.mark.parametrize(
    "max_lines_per_cell, cell_lines, expected_ret, expected_pass", [
        (3, [0, 1, 2, 3, 4], [LR(x, LintType.LINES_PER_CELL) for x in [True, True, True, True, False]], False),
        (3, [0, 1, 2, 3], [LR(x, LintType.LINES_PER_CELL) for x in [True, True, True, True]], True),
        (-1, [0, 1], [], True),
    ]
)
def test_lines_per_cell(max_lines_per_cell, cell_lines, expected_ret, expected_pass):
    ret, passed = lint_lines_per_cell(cell_lines, max_lines_per_cell)
    _verify(ret, passed, expected_ret, expected_pass)


@pytest.mark.parametrize(
    "max_cells_per_notebook, cell_count, expected_ret, expected_pass", [
        (5, 10, [LR(False, LintType.CELLS_PER_NOTEBOOK)], False),
        (5, 3, [LR(True, LintType.CELLS_PER_NOTEBOOK)], True),
        (0, 10, [LR(False, LintType.CELLS_PER_NOTEBOOK)], False),
        (10, 0, [LR(True, LintType.CELLS_PER_NOTEBOOK)], True),
        (-1, 10, [], True)
    ]
)
def test_cells_per_notebook(max_cells_per_notebook, cell_count, expected_ret, expected_pass):
    ret, passed = lint_cells_per_notebook(cell_count, max_cells_per_notebook)
    _verify(ret, passed, expected_ret, expected_pass)


@pytest.mark.parametrize(
    "max_function_definitions, functions, expected_ret, expected_pass", [
        (5, 10, [LR(False, LintType.FUNCTION_DEFINITIONS)], False),
        (5, 3, [LR(True, LintType.FUNCTION_DEFINITIONS)], True),
        (0, 10, [LR(False, LintType.FUNCTION_DEFINITIONS)], False),
        (10, 0, [LR(True, LintType.FUNCTION_DEFINITIONS)], True),
        (-1, 10, [], True)
    ]
)
def test_lint_function_definitions(max_function_definitions, functions, expected_ret, expected_pass):
    ret, passed = lint_function_definitions(functions, max_function_definitions)
    _verify(ret, passed, expected_ret, expected_pass)


@pytest.mark.parametrize(
    "max_class_definitions, classes, expected_ret, expected_pass", [
        (5, 10, [LR(False, LintType.CLASS_DEFINITIONS)], False),
        (5, 3, [LR(True, LintType.CLASS_DEFINITIONS)], True),
        (0, 10, [LR(False, LintType.CLASS_DEFINITIONS)], False),
        (10, 0, [LR(True, LintType.CLASS_DEFINITIONS)], True),
        (-1, 10, [], True)
    ]
)
def test_lint_class_definitions(max_class_definitions, classes, expected_ret, expected_pass):
    ret, passed = lint_class_definitions(classes, max_class_definitions)
    _verify(ret, passed, expected_ret, expected_pass)


@pytest.mark.parametrize(
    "kernelspec_requirements, kernelspec, expected_ret, expected_pass", [
        ({'name': 'python3'}, {'name': 'python3',
                               'display_name': 'Python 3'}, [LR(True, LintType.KERNELSPEC)], True),
        ({'name': 'python3'}, {'name': 'python2'}, [LR(False, LintType.KERNELSPEC)], False),
        ({}, {'name': 'python3',
              'display_name': 'Python 3'}, [LR(True, LintType.KERNELSPEC)], True),
        (False, {}, [], True),
        (False, {'something': 'else'}, [], True),
    ]
)
def test_kernelspec(kernelspec_requirements, kernelspec, expected_ret, expected_pass):
    ret, passed = lint_kernelspec(kernelspec, kernelspec_requirements)
    _verify(ret, passed, expected_ret, expected_pass)


@pytest.mark.parametrize(
    "magics_whitelist, magics_blacklist, magics, expected_ret, expected_pass", [
        # no check
        (None, None, [], [], True),
        (None, None, ['anything'], [], True),
        # empty whitelist: no magics allowed
        ([], None, ['anything'], [LR(False, LintType.MAGICS)], False),
        # only whitelisted
        (['ok1', 'ok2'], None, ['ok1'], [LR(True, LintType.MAGICS)], True),
        (['ok1', 'ok2'], None, ['notok'], [LR(False, LintType.MAGICS)], False),
        # no blacklisted
        (None, ['notok'], ['ok'], [LR(True, LintType.MAGICS)], True),
        (None, ['notok'], ['notok'], [LR(False, LintType.MAGICS)], False),
    ]
)
def test_magics(magics_whitelist, magics_blacklist, magics, expected_ret, expected_pass):
    ret, passed = lint_magics(magics, whitelist=magics_whitelist, blacklist=magics_blacklist)
    _verify(ret, passed, expected_ret, expected_pass)


def test_magics_lists_sanity():
    msg = "Must specify either a whitelist or a blacklist, not both."

    with pytest.raises(ValueError, match=msg):
        lint_magics(set(), whitelist=['one'], blacklist=['one'])

    with pytest.raises(ValueError, match=msg):
        lint_magics(set(), whitelist=[], blacklist=['one'])

    with pytest.raises(ValueError, match=msg):
        lint_magics(set(), whitelist=[], blacklist=[])

    with pytest.raises(ValueError, match=msg):
        lint_magics(set(), whitelist=['one'], blacklist=[])


@pytest.mark.parametrize(
    "rules, noqa_regex, expected_ret, expected_pass", [
        # no rules
        ({}, None, [], True),
        # one rule, pass
        ({'lines_per_cell': -1}, None, [], True),
        # one rule, fail
        ({'lines_per_cell': 1}, None, [LR(x, LintType.LINES_PER_CELL) for x in [False, True, False, False]], False),
        # multiple rules, combo fail
        ({'lines_per_cell': 5,
          'cells_per_notebook': 1}, None, [LR(True, LintType.LINES_PER_CELL)] * 4 + [LR(False, LintType.CELLS_PER_NOTEBOOK)], False),
        # multiple rules, one disabled, combo pass
        ({'lines_per_cell': 1,
          'cells_per_notebook': 10}, "# LINT_DISABLE: notebook_(.*)$", [LR(True, LintType.CELLS_PER_NOTEBOOK)], True),
        # multiple rules, combo pass
        ({'lines_per_cell': 5,
          'cells_per_notebook': 10}, None, [LR(True, LintType.LINES_PER_CELL)] * 4 + [LR(True, LintType.CELLS_PER_NOTEBOOK)], True),
        # all the expected rules
        ({'lines_per_cell': 5,
          'cells_per_notebook': 2,
          'function_definitions': 0,
          'class_definitions': 0,
          'kernelspec_requirements':
            {'name': 'python3'},
          'magics_whitelist': ['matplotlib']}, None, [LR(True, LintType.LINES_PER_CELL)] * 4 +
                                                     [LR(False, LintType.CELLS_PER_NOTEBOOK)] +
                                                     [LR(False, LintType.FUNCTION_DEFINITIONS)] +
                                                     [LR(False, LintType.CLASS_DEFINITIONS)] +
                                                     [LR(True, LintType.KERNELSPEC)] +
                                                     [LR(True, LintType.MAGICS)], False)
    ]
)
def test_run(rules, noqa_regex, expected_ret, expected_pass):
    nb = os.path.join(os.path.dirname(__file__), 'more.ipynb')
    ret, passed = run(nb, rules=rules, noqa_regex=noqa_regex)
    _verify(ret, passed, expected_ret, expected_pass)


def _verify(ret, passed, expected_ret, expected_pass):
    assert [(r.passed, r.type) for r in ret] == [(e.passed, e.type) for e in expected_ret]
    assert passed is expected_pass


def test_runWithHTMLReturn_norules():
    # no rules, so will pass without any checks
    nb = os.path.join(os.path.dirname(__file__), 'more.ipynb')
    html, passed = runWithHTMLReturn(nb)
    assert passed is True
    _check(html, [])


def test_runWithHTMLReturn_pass():
    # checks should pass
    nb = os.path.join(os.path.dirname(__file__), 'more.ipynb')
    html, passed = runWithHTMLReturn(nb, rules={"cells_per_notebook": 10})
    assert passed is True
    _check(html, [("PASSED", "Checking cells per notebook (max=10; actual=4) (Notebook)")])


def test_runWithHTMLReturn_fail():
    # checks should fail
    nb = os.path.join(os.path.dirname(__file__), 'more.ipynb')
    html, passed = runWithHTMLReturn(nb, rules={"cells_per_notebook": 1, "lines_per_cell": 2})
    assert passed is False
    _check(html, [
        ("PASSED", "Checking lines in cell (max=2; actual=2)(Cell 1)"),
        ("PASSED", "Checking lines in cell (max=2; actual=1)(Cell 2)"),
        ("PASSED", "Checking lines in cell (max=2; actual=2)(Cell 3)"),
        ("FAILED", "Checking lines in cell (max=2; actual=3)(Cell 4)"),
        ("FAILED", "Checking cells per notebook (max=1; actual=4) (Notebook)")])


def _check(html, expected_results):
    # quick checking html matches expected results
    soup = BeautifulSoup(html, "html.parser")
    actual_results = soup.find_all("p")
    assert len(actual_results) == len(expected_results)
    for actual, expected in zip(sorted(actual_results, key=lambda x: x.text[x.text.find("Checking ")::]), sorted(expected_results, key=itemgetter(1))):
        assert actual.text.startswith(expected[0])
        assert actual.text.endswith(expected[1])



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import json
import os
import os.path
import nbformat
import sys
import tornado.gen
from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor
from notebook.base.handlers import IPythonHandler
from notebook.utils import url_path_join
try:
    from backports.tempfile import TemporaryDirectory
except ImportError:
    from tempfile import TemporaryDirectory

from .test import runWithHTMLReturn as runTest
from .lint import runWithHTMLReturn as runLint


class RunCelltestsHandler(IPythonHandler):
    executor = ThreadPoolExecutor(4)

    def initialize(self, rules=None, executable=None):
        self.rules = rules
        self.executable = executable

    def get(self):
        self.finish({'status': 0, 'test': self.rules})

    @run_on_executor
    def _run(self, body, path, name):
        with TemporaryDirectory() as tempdir:
            path = os.path.abspath(os.path.join(tempdir, name))
            node = nbformat.from_dict(body.get('model'))
            nbformat.write(node, path)
            ret = runTest(path, executable=self.executable, rules=self.rules)
            return ret

    @tornado.gen.coroutine
    def post(self):
        body = json.loads(self.request.body)
        path = os.path.join(os.getcwd(), body.get('path'))
        name = os.path.basename(path)
        ret = yield self._run(body, path, name)
        self.finish({'status': 0, 'test': ret})


class RunLintsHandler(IPythonHandler):
    executor = ThreadPoolExecutor(4)

    def initialize(self, rules=None, executable=None):
        self.rules = rules
        self.executable = executable

    def get(self):
        self.finish({'status': 0, 'linters': self.rules})

    @run_on_executor
    def _run(self, body, path, name):
        with TemporaryDirectory() as tempdir:
            path = os.path.abspath(os.path.join(tempdir, name))
            node = nbformat.from_dict(body.get('model'))
            nbformat.write(node, path)
            ret, status = runLint(path, executable=self.executable, rules=self.rules)
            return ret, status
            self.finish({'status': status, 'lint': ret})

    @tornado.gen.coroutine
    def post(self):
        body = json.loads(self.request.body)
        path = os.path.join(os.getcwd(), body.get('path'))
        name = os.path.basename(path)
        ret, status = yield self._run(body, path, name)
        self.finish({'status': status, 'lint': ret})


def _load_jupyter_server_extension(nb_server_app):
    """
    Called when the extension is loaded.

    Args:
        nb_server_app (NotebookWebApplication): handle to the Notebook webserver instance.
    """
    web_app = nb_server_app.web_app

    host_pattern = '.*$'
    base_url = web_app.settings['base_url']
    # host_pattern = '.*$'
    print('Installing nbcelltests handler on path %s' % url_path_join(base_url, 'celltests'))

    rules = nb_server_app.config.get('JupyterLabCelltests', {}).get('rules', {})
    test_executable = nb_server_app.config.get('JupyterLabCelltests', {}).get('test_executable', [sys.executable, '-m', 'pytest', '-v'])
    lint_executable = nb_server_app.config.get('JupyterLabCelltests', {}).get('lint_executable', [sys.executable, '-m', 'flake8', '--ignore=W391'])

    web_app.add_handlers(host_pattern, [(url_path_join(base_url, 'celltests/test/run'), RunCelltestsHandler, {'rules': rules, 'executable': test_executable})])
    web_app.add_handlers(host_pattern, [(url_path_join(base_url, 'celltests/lint/run'), RunLintsHandler, {'rules': rules, 'executable': lint_executable})])



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import json
import nbformat
import os
import shutil
import sys
import subprocess
import tempfile
from .define import TestMessage, TestType
from .shared import extract_extrametadata, get_coverage
from .tests_vendored import BASE, JSON_CONFD


def run(notebook, rules=None, filename=None, kernel_name="", current_env=False):
    """
    Runs no tests: just generates test script for supplied notebook.

    kernel_name and current_env 'will be passed to nbval'.
    """
    nb = nbformat.read(notebook, 4)
    name = filename or os.path.splitext(notebook)[0] + '_test.py'
    extra_metadata = extract_extrametadata(nb)
    rules = rules or {}
    extra_metadata.update(rules)

    # TODO: Coverage shouldn't be recorded at generation time as it
    # will go stale if the notebook changes. Should move to same
    # mechanism as source/tests. However, we plan to replace coverage
    # with code coverage measured during test execution.
    coverage = []
    if 'cell_coverage' in extra_metadata:
        coverage.append((get_coverage(extra_metadata), extra_metadata['cell_coverage']))

    # output tests to test file
    with open(name, 'w', encoding='utf-8') as fp:
        fp.write(BASE.format(kernel_name=kernel_name, current_env=current_env, path_to_notebook=notebook, coverage=coverage))

    return name


def runWithReturn(notebook, executable=None, **run_kw):
    """
    Run notebook's celltests in a subprocess and return exit status.

    run_kw will be passed to run().
    """
    name = run(notebook, **run_kw)
    executable = executable or [sys.executable, '-m', 'pytest', '-v']
    argv = executable + [name]
    return subprocess.check_output(argv)


def _pytest_nodeid_prefix(path):
    return os.path.splitdrive(path)[1][1:].replace(os.path.sep, '/') + '/'


def runWithReport(notebook, executable=None, collect_only=False, **run_kw):
    tmpd = tempfile.mkdtemp()
    py_file = os.path.join(tmpd, os.path.basename(notebook).replace('.ipynb', '.py'))
    json_file = os.path.join(tmpd, os.path.basename(notebook).replace('.ipynb', '.json'))
    _ = run(notebook, filename=py_file, **run_kw)
    ret = []
    try:
        # enable collecting info via json
        with open(os.path.join(tmpd, 'conftest.py'), 'w', encoding='utf8') as f:
            f.write(JSON_CONFD)
        executable = executable or [sys.executable, '-m', 'pytest', '-v']
        argv = executable + ['--internal-json-report=' + json_file, py_file]
        if collect_only:
            argv.append('--collect-only')
        subprocess.call(argv)
        with open(json_file, 'r', encoding='utf-8') as f:
            # load json from file
            data = json.load(f)

        data = [d for d in data['reports'] + data['collected_items'] if d.get('nodeid', '') and (collect_only or d.get('when') == 'call')]
        prefix = _pytest_nodeid_prefix(tmpd)
        for node in data:
            # replace relative path, strip leading ::
            node['nodeid'] = node['nodeid'].replace(prefix, '').strip(':')

            if node.get('outcome') == 'passed':
                outcome = 1
            elif 'outcome' not in node:
                outcome = 0
            else:
                outcome = -1

            if 'test_cell_coverage' in node['nodeid']:
                ret.append(TestMessage(-1, 'Testing cell coverage', TestType.CELL_COVERAGE, outcome))
            elif 'test_cell' in node['nodeid']:
                cell_no = node['nodeid'].rsplit('_', 1)[-1]
                ret.append(TestMessage(int(cell_no) + 1, 'Testing cell', TestType.CELL_TEST, outcome))
            else:
                continue
    finally:
        shutil.rmtree(tmpd)
    return ret


def runWithHTMLReturn(notebook, executable=None, **run_kw):
    """
    Run notebook's celltests in a subprocess and return html generated
    by pytest's --self-contained-html.

    rules: coverage requirements (if any).

    Note - leaves behind the following generated files for
    "/path/to/notebook.ipynb":
      * /path/to/notebook_test.py (notebook test script)
      * /path/to/notebook_test.html (pytest's html report)
    """
    name = run(notebook, **run_kw)
    html = name.replace('.py', '.html')
    executable = executable or [sys.executable, '-m', 'pytest', '-v']
    argv = executable + ['--html=' + html, '--self-contained-html', name]
    subprocess.call(argv)
    with open(html, 'r', encoding='utf-8') as fp:
        return fp.read()


def runWithHTMLReturn2(notebook, executable=None, **run_kw):
    '''use custom return objects'''
    ret = ''
    executable = executable or [sys.executable, '-m', 'pytest', '-v']
    ret_tmp = run(notebook, **run_kw)
    for test in ret_tmp:
        test = test.to_html()
        ret += '<p>' + test + '</p>'
    return '<div style="display: flex; flex-direction: column;">' + test + '</div>'


if __name__ == '__main__':
    if len(sys.argv) != 2:
        raise Exception('Usage:python -m nbcelltests.test <ipynb file>')
    notebook = sys.argv[1]
    # TODO: seems likely this should use one of the above run fns.
    name = run(notebook)
    argv = [sys.executable, '-m', 'pytest', name, '-v', '--html=' + name.replace('.py', '.html'), '--self-contained-html']
    print(' '.join(argv))
    subprocess.call(argv)



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
from ._version import __version__  # noqa: F401


def _jupyter_server_extension_paths():
    return [{
        "module": "nbcelltests"
    }]


def load_jupyter_server_extension(nb_server_app):
    # avoid pulling in extension whenever nbcelltests is imported (e.g. for cli)
    from .extension import _load_jupyter_server_extension
    _load_jupyter_server_extension(nb_server_app)



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import sys
import ast
import re
import nbconvert


# note: could consider combining these separate classes
class FnDefCounter(ast.NodeVisitor):
    """Counts function definitions (ignoring class methods)."""

    count = 0

    def visit_FunctionDef(self, node):
        self.count += 1

    def visit_ClassDef(self, node):
        return

    # to count lambdas, add visit_Lambda


class ClassDefCounter(ast.NodeVisitor):
    """Counts class declarations."""

    count = 0

    def visit_ClassDef(self, node):
        self.count += 1


class MagicsRecorder(ast.NodeVisitor):
    """Record magics.

    Magics are like this:
      * get_ipython().run_line_magic(name, line)
      * get_ipython().run_cell_magic(name, line, cell)

    Or this (py2):
      * get_ipython().magic(magic_string)

    """

    magic_fn_names_py3 = set(['run_line_magic',
                              'run_cell_magic'])
    magic_fn_names_py2 = set(['magic'])

    magic_fn_names = magic_fn_names_py3 | magic_fn_names_py2

    def __init__(self):
        self.seen = set()

    def visit_Call(self, node):
        if hasattr(node.func, 'attr') and node.func.attr in self.magic_fn_names:
            # should maybe find ipython's own parsing code and use that instead
            if node.func.value.func.id == 'get_ipython':
                magic_name = node.args[0].s
                if node.func.attr in self.magic_fn_names_py2:
                    magic_name = magic_name.split()[0]  # (again, find ipython's parsing?)
                self.seen.add(magic_name)


# Note: I think it's confusing to insert the actual counts into the
# metadata.  Why not keep them separate?
#
# Note: this always does everything, which might be unnecessary
# (e.g. if haven't asked for magics checking, don't need to extract
# them)
def extract_extrametadata(notebook, override=None, noqa_regex=None):
    if noqa_regex is not None:
        noqa_regex = re.compile(noqa_regex)
        if not noqa_regex.groups == 1:
            raise ValueError("noqa_regex must contain one capture group (specifying the rule)")

    base = notebook.metadata.get('celltests', {})
    override = override or {}
    base['lines'] = 0  # TODO: is this used?
    base['kernelspec'] = notebook.metadata.get('kernelspec', {})

    # "python code" things (e.g. number of function definitions)...
    # note: no attempt to be clever here (so e.g. "%time def f: pass" would be missed, as would the contents of
    # a cell using %%capture cell magics; possible to handle those scenarios but would take more effort)
    code = nbconvert.PythonExporter(exclude_raw=True).from_notebook_node(notebook)[0]
    # notebooks start with "coding: utf-8"
    parsed_source = ast.parse(code.encode('utf8') if sys.version_info[0] == 2 else code)

    fn_def_counter = FnDefCounter()
    fn_def_counter.visit(parsed_source)
    base['functions'] = fn_def_counter.count

    class_counter = ClassDefCounter()
    class_counter.visit(parsed_source)
    base['classes'] = class_counter.count

    # alternative to doing it this way would be to check the ipython
    # souce for %magic, %%magics before it's converted to regular python
    magics_recorder = MagicsRecorder()
    magics_recorder.visit(parsed_source)
    base['magics'] = magics_recorder.seen

    # "notebook structure" things...
    base['cell_count'] = 0
    base['cell_tested'] = []
    base['test_count'] = 0
    base['cell_lines'] = []
    base['noqa'] = set()

    for c in notebook.cells:
        if c['cell_type'] != 'code' or empty_ast(c['source']):
            continue

        base['cell_lines'].append(0)
        base['cell_tested'].append(False)
        base['cell_count'] += 1

        for line in c['source'].split('\n'):
            if noqa_regex:
                noqa_match = noqa_regex.match(line)
                if noqa_match:
                    base['noqa'].add(noqa_match.group(1))
            if not empty_ast(line):
                base['lines'] += 1
                base['cell_lines'][-1] += 1
        if cell_injected_into_test(get_test(c)):
            base['test_count'] += 1
            base['cell_tested'][-1] = True

    # in case you want to override notebook settings
    if override:
        base.update(override)

    return base


def get_test(cell):
    return lines2source(cell.get('metadata', {}).get('tests', []))


def empty_ast(source):
    """
    Whether the supplied source string has an empty ast.

    >>> empty_ast(" ")
    True

    >>> empty_ast("pass")
    False

    >>> empty_ast("# hello")
    True

    """
    try:
        parsed = ast.parse(source)
    except SyntaxError:
        # If there's a syntax error, it's not an empty code cell.
        # Handling and communicating syntax errors is a general issue
        # (https://github.com/jpmorganchase/nbcelltests/issues/101).
        # Note: this will also handle magics.
        return False

    # TODO: py2 utf8
    return len(parsed.body) == 0


def only_whitespace(source):
    """
    Whether the supplied source string contains only whitespace.

    >>> only_whitespace(" ")
    True

    >>> only_whitespace("pass")
    False

    >>> only_whitespace("# hello")
    False
    """
    return len(source.strip()) == 0


# TODO drop the "token" part
CELL_INJ_TOKEN = r"%cell"
CELL_SKIP_TOKEN = r"# no %cell"


def source2lines(source):
    return source.splitlines(keepends=True)


def lines2source(lines):
    return "".join(lines)


def get_cell_inj_span(test_line):
    """
    Return the location of %cell in the given line as (start_index,
    end_index), or None if %cell does not occur.
    """
    if not test_line.strip().startswith(CELL_INJ_TOKEN):
        return None
    else:
        cell_start = test_line.index(CELL_INJ_TOKEN)
        cell_end = cell_start + len(CELL_INJ_TOKEN)
        return cell_start, cell_end


def cell_injected_into_test(test_source):
    """
    Return True if the corresponding cell is injected into the test,
    False if the cell is deliberately not injected, or None if there
    is no deliberate command either way.
    """
    inject = False
    run = None
    for test_line in source2lines(test_source):
        if inject is False and get_cell_inj_span(test_line) is not None:
            inject = True
        elif run is None and test_line.strip().startswith(CELL_SKIP_TOKEN):
            run = False

    if run is False and inject:
        raise ValueError("'%s' and '%s' are mutually exclusive but both were supplied:\n%s" % (CELL_SKIP_TOKEN, CELL_INJ_TOKEN, test_source))

    return inject or run


def get_coverage(metadata):
    if metadata['cell_count'] == 0:
        return 0
    return 100.0 * metadata['test_count'] / metadata['cell_count']



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
from enum import Enum


class LintType(Enum):
    LINES_PER_CELL = 'lines_per_cell'
    CELLS_PER_NOTEBOOK = 'cells_per_notebook'
    FUNCTION_DEFINITIONS = 'function_definitions'
    CLASS_DEFINITIONS = 'class_definitions'
    LINTER = 'linter'
    KERNELSPEC = 'kernelspec'
    MAGICS = 'magics'


class TestType(Enum):
    CELL_COVERAGE = 'cell_coverage'
    CELL_TEST = 'cell_test'


class LintMessage(object):
    def __init__(self, cell, message, type, passed=False):
        self.cell = cell
        self.message = message
        self.type = type
        self.passed = passed

    def __repr__(self):
        ret = 'PASSED: ' if self.passed else 'FAILED: '
        ret += self.message
        ret += " (Cell %d)" % self.cell if self.cell > 0 else " (Notebook)"
        return ret

    def to_html(self):
        ret = '<span style="color: green;">PASSED&nbsp;</span>' if self.passed else '<span style="color: red;">FAILED&nbsp;</span>'
        ret += self.message
        ret += "(Cell %d)" % self.cell if self.cell > 0 else " (Notebook)"
        return ret


class TestMessage(object):
    def __init__(self, cell, message, type, passed=0):
        self.cell = cell
        self.message = message
        self.type = type
        self.passed = passed

    def __repr__(self):
        ret = 'PASSED: ' if self.passed > 0 else 'FAILED: ' if self.passed < 0 else 'NOT RUN: '
        ret += self.message
        ret += " (Cell %d)" % self.cell if self.cell > 0 else " (Notebook)"
        return ret

    def to_html(self):
        ret = '<span style="color: green;">PASSED&nbsp;</span>' if self.passed else '<span style="color: red;">FAILED&nbsp;</span>'
        ret += self.message
        ret += "(Cell %d)" % self.cell if self.cell > 0 else " (Notebook)"
        return ret



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import nbformat
import os
import sys
import subprocess
from nbconvert import ScriptExporter
from tempfile import NamedTemporaryFile
from .shared import extract_extrametadata
from .define import LintMessage, LintType


def lint_lines_per_cell(cell_lines, max_lines_per_cell=-1):
    ret = []
    if max_lines_per_cell < 0:
        return [], True
    for i, lines_in_cell in enumerate(cell_lines):
        ret.append(
            LintMessage(
                i + 1,  # TODO: ambiguous - e.g. cell 0 or first cell?
                'Checking lines in cell (max={max_}; actual={actual})'.format(
                    max_=max_lines_per_cell,
                    actual=lines_in_cell),
                LintType.LINES_PER_CELL,
                lines_in_cell <= max_lines_per_cell))
    return ret, all([x.passed for x in ret])


def lint_cells_per_notebook(cell_count, max_cells_per_notebook=-1):
    if max_cells_per_notebook < 0:
        return [], True
    passed = cell_count <= max_cells_per_notebook
    return [LintMessage(-1, 'Checking cells per notebook (max={max_}; actual={actual})'.format(max_=max_cells_per_notebook, actual=cell_count), LintType.CELLS_PER_NOTEBOOK, passed)], passed


def lint_function_definitions(functions, max_function_definitions=-1):
    if max_function_definitions < 0:
        return [], True
    passed = functions <= max_function_definitions
    return [LintMessage(-1, 'Checking functions per notebook (max={max_}; actual={actual})'.format(max_=max_function_definitions, actual=functions), LintType.FUNCTION_DEFINITIONS, passed)], passed


def lint_class_definitions(classes, max_class_definitions=-1):
    if max_class_definitions < 0:
        return [], True
    passed = classes <= max_class_definitions
    return [LintMessage(-1, 'Checking classes per notebook (max={max_}; actual={actual})'.format(max_=max_class_definitions, actual=classes), LintType.CLASS_DEFINITIONS, passed)], passed


def lint_kernelspec(kernelspec, kernelspec_requirements=False):
    """Check that kernelspec fulfills kernelspec_requirements.

    If kernelspec_requirements is False, no check will happen.

    If kernelspec_requirements is None, requires an empty kernelspec
    (use to enforce saving without kernelspec details).

    Otherwise, kernelspec must contain at least the same key: value
    pairs as are in kernelspec_requirements.
    """
    if kernelspec_requirements is False:
        return [], True
    # assumes kernelspec dict values are hashable (they're strings)
    passed = set(kernelspec.items()).issuperset(kernelspec_requirements.items())
    return [LintMessage(-1, 'Checking kernelspec (min. required={required}; actual={actual})'.format(required=kernelspec_requirements, actual=kernelspec), LintType.KERNELSPEC, passed)], passed


def lint_magics(magics, whitelist=None, blacklist=None):
    """Check that magics are acceptable.

    Specify either a whitelist or a blacklist (or neither), but not
    both.
    """
    if whitelist is None and blacklist is None:
        return [], True

    if whitelist is not None and blacklist is not None:
        raise ValueError("Must specify either a whitelist or a blacklist, not both. Blacklist: {}; whitelist: {}".format(blacklist, whitelist))

    if whitelist is not None:
        bad = set(magics) - set(whitelist)
        msg = "missing from whitelist:"
    elif blacklist is not None:
        bad = set(magics) & set(blacklist)
        msg = "present in blacklist:"

    passed = not(bad)
    return [LintMessage(-1, 'Checking magics{}'.format(" ({} {})".format(msg, bad) if bad else ""), LintType.MAGICS, passed)], passed


def run(notebook, executable=None, rules=None, noqa_regex=None):
    nb = nbformat.read(notebook, 4)
    extra_metadata = extract_extrametadata(nb, noqa_regex=noqa_regex)
    ret = []
    passed = True

    rules = rules or {}
    extra_metadata.update(rules)

    # TODO: consider warning if referring to non-existent rules
    rules_to_remove = extra_metadata['noqa'] & extra_metadata.keys()
    for rule in rules_to_remove:
        del extra_metadata[rule]

    # TODO: lintfail is more like lintpassed?

    if 'lines_per_cell' in extra_metadata:
        lintret, lintfail = lint_lines_per_cell(extra_metadata['cell_lines'], max_lines_per_cell=extra_metadata['lines_per_cell'])
        ret.extend(lintret)
        passed = passed and lintfail

    if 'cells_per_notebook' in extra_metadata:
        lintret, lintfail = lint_cells_per_notebook(extra_metadata['cell_count'], max_cells_per_notebook=extra_metadata['cells_per_notebook'])
        ret.extend(lintret)
        passed = passed and lintfail

    if 'function_definitions' in extra_metadata:
        lintret, lintfail = lint_function_definitions(extra_metadata['functions'], max_function_definitions=extra_metadata['function_definitions'])
        ret.extend(lintret)
        passed = passed and lintfail

    if 'class_definitions' in extra_metadata:
        lintret, lintfail = lint_class_definitions(extra_metadata['classes'], max_class_definitions=extra_metadata['class_definitions'])
        ret.extend(lintret)
        passed = passed and lintfail

    if 'kernelspec_requirements' in extra_metadata:
        lintret, lintfail = lint_kernelspec(kernelspec=extra_metadata['kernelspec'], kernelspec_requirements=extra_metadata['kernelspec_requirements'])
        ret.extend(lintret)
        passed = passed and lintfail

    if 'magics_whitelist' in extra_metadata or 'magics_blacklist' in extra_metadata:
        lintret, lintfail = lint_magics(magics=extra_metadata['magics'], whitelist=extra_metadata.get('magics_whitelist', None), blacklist=extra_metadata.get('magics_blacklist', None))
        ret.extend(lintret)
        passed = passed and lintfail

    if executable:
        exp = ScriptExporter()
        (body, resources) = exp.from_notebook_node(nb)
        tf = NamedTemporaryFile(mode='w', suffix='.py', delete=False, encoding='utf8')
        tf_name = tf.name
        try:
            tf.write(body)
            tf.close()
            executable.append(tf_name)
            ret2 = _run_and_capture_utf8(executable)
            msg = ret2.stdout + '\t' + ret2.stderr
            ret.append(LintMessage(-1, 'Checking lint:\n' + msg.strip(), LintType.LINTER, False if msg.strip() else True))
        finally:
            os.remove(tf_name)

    return ret, passed


def _run_and_capture_utf8(args):
    # PYTHONIOENCODING for pyflakes on Windows
    run_kw = {'env': dict(os.environ, PYTHONIOENCODING='utf8')} if sys.platform == 'win32' else {}
    return subprocess.run(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8', **run_kw)


def runWithHTMLReturn(notebook, executable=None, rules=None):
    ret = ''
    ret_tmp, fail = run(notebook, executable=executable, rules=rules)
    for lint in ret_tmp:
        lint = lint.to_html()
        ret += '<p>' + lint + '</p>'
    return '<div style="display: flex; flex-direction: column;">' + ret + '</div>', fail


if __name__ == '__main__':
    # TODO: doesn't support the typical interface of run (e.g. rules)
    if len(sys.argv) != 2:
        raise Exception('Usage:python -m nbcelltests.lint <ipynb file>')
    notebook = sys.argv[1]
    ret, passed = run(notebook, ['flake8', '--ignore=W391'])
    if passed:
        print('\n'.join(str(r) for r in ret))
        sys.exit(0)
    else:
        print('\n'.join(str(r) for r in ret))
        sys.exit(1)



# *****************************************************************************
#
# Copyright (c) 2019, the nbcelltests authors.
#
# This file is part of the nbcelltests library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import os

import pytest
import nbformat

from nbcelltests.shared import extract_extrametadata, get_coverage, empty_ast, only_whitespace, cell_injected_into_test, get_cell_inj_span

# TODO: should generate these
BASIC_NB = os.path.join(os.path.dirname(__file__), 'basic.ipynb')
MORE_NB = os.path.join(os.path.dirname(__file__), 'more.ipynb')
MAGICS_NB = os.path.join(os.path.dirname(__file__), 'magics.ipynb')
COVERAGE_NB = os.path.join(os.path.dirname(__file__), '_cell_coverage.ipynb')
LINT_DISABLE_NB = os.path.join(os.path.dirname(__file__), '_lint_disable.ipynb')

# TODO should parameterize test_empty_ast _whitespace

# empty ast, empty whitespace
_empty_empty = [
    "",
    " ",
    "\t",
    "\n",
    "\r\n"
]

# empty ast, non-empty whitespace
_empty_nonempty = [
    "#pass",
]


# non-empty ast, non-empty whitespace
_nonempty_nonempty = [
    "import blah\nblah.do_something()",
    "%matplotlib inline",
    "pass"
]


def test_empty_ast():
    for x in _empty_empty:
        assert empty_ast(x) is True

    for x in _empty_nonempty:
        assert empty_ast(x) is True

    for x in _nonempty_nonempty:
        assert empty_ast(x) is False


def test_only_whitespace():
    for x in _empty_empty:
        assert only_whitespace(x) is True

    for x in _empty_nonempty:
        assert only_whitespace(x) is False

    for x in _nonempty_nonempty:
        assert only_whitespace(x) is False


def _metadata(nb, what):
    extra_metadata = extract_extrametadata(nbformat.read(nb, 4))
    return extra_metadata[what]


def test_extract_extrametadata_functions_basic():
    assert _metadata(BASIC_NB, 'functions') == 2


def test_extract_extrametadata_classes_basic():
    assert _metadata(BASIC_NB, 'classes') == 2


def test_extract_extrametadata_cell_count_basic():
    assert _metadata(BASIC_NB, 'cell_count') == 5


def test_extract_extrametadata_cell_lines_basic():
    assert _metadata(BASIC_NB, 'cell_lines') == [1] * 5

# with some magics present


def test_extract_extrametadata_functions_more():
    assert _metadata(MORE_NB, 'functions') == 1


def test_extract_extrametadata_classes_more():
    assert _metadata(MORE_NB, 'classes') == 1


def test_extract_extrametadata_cell_count_more():
    assert _metadata(MORE_NB, 'cell_count') == 4


def test_extract_extrametadata_cell_lines_more():
    assert _metadata(MORE_NB, 'cell_lines') == [2, 1] + [2, 3]


def test_extract_extrametadata_magics():
    assert _metadata(MAGICS_NB, 'magics') == set(['magics1', 'magics2', 'magics3'])


# with non-code cells present


def test_extract_extrametadata_functions_noncode():
    assert _metadata(COVERAGE_NB, 'functions') == 1


def test_extract_extrametadata_classes_noncode():
    assert _metadata(COVERAGE_NB, 'classes') == 1


def test_extract_extrametadata_cell_count_noncode():
    assert _metadata(COVERAGE_NB, 'cell_count') == 4


def test_extract_extrametadata_cell_lines_noncode():
    assert _metadata(COVERAGE_NB, 'cell_lines') == [3, 1, 1, 2]


def test_extract_extrametadata_magics_noncode():
    assert _metadata(COVERAGE_NB, 'magics') == set(['dirs'])


# coverage

# tests would clearer with multiple notebooks containing independent
# cases - i.e. should generate them

# N = contribs to numerator of calc
# D = contribs to denominator of calc
# - = contribs nothing

# --  empty code cell with test
# --  markdown cell
# ND  code cell with valid test
# --  raw cell
# -D  code cell with invalid test (no %cell)
# -D  code cell with empty test
# -D  code cell with no test
#
# i.e. coverage = 1/4
def test_extract_extrametadata_cell_count_coverage():
    assert _metadata(COVERAGE_NB, 'cell_count') == 4


def test_extract_extrametadata_cell_test_coverage():
    assert _metadata(COVERAGE_NB, 'test_count') == 1


def test_get_coverage_nb():
    assert get_coverage(extract_extrametadata(nbformat.read(COVERAGE_NB, 4))) == 25


@pytest.mark.parametrize(
    "test_count, cell_count, expected", [
        (0, 10, 0),
        (5, 10, 50),
        (10, 10, 100),
        (0, 0, 0),
        (10, 0, 0),  # this would be an error at test generation time
    ]
)
def test_get_coverage(test_count, cell_count, expected):
    metadata = {
        'test_count': test_count,
        'cell_count': cell_count
    }
    assert get_coverage(metadata) == expected


# lint disable

def test_extract_extrametadata_disable_none():
    metadata = extract_extrametadata(nbformat.read(LINT_DISABLE_NB, 4))
    assert len(metadata['noqa']) == 0


def test_extract_extrametadata_disable_notpresent():
    metadata = extract_extrametadata(nbformat.read(LINT_DISABLE_NB, 4), noqa_regex=r"^# don't noqa notebook:\s*(.*)$")
    assert len(metadata['noqa']) == 0


def test_extract_extrametadata_disable_cells_count():
    metadata = extract_extrametadata(nbformat.read(LINT_DISABLE_NB, 4), noqa_regex=r"^# noqa notebook:\s*(.*)$")
    assert metadata['noqa'] == {'cells_per_notebook'}


def test_extract_extrametadata_disable_bad_regex():
    try:
        metadata = extract_extrametadata(nbformat.read(LINT_DISABLE_NB, 4), noqa_regex=r"^# noqa notebook:\s*.*$")
    except ValueError as e:
        assert e.args[0] == "noqa_regex must contain one capture group (specifying the rule)"
    else:
        assert False, "should have raised a ValueError"


@pytest.mark.parametrize(
    "test_line, expected", [
        (r"%cell", (0, 5)),
        (r"%cell;x", (0, 5)),
        pytest.param(r"%celll", None, marks=pytest.mark.xfail(reason="need to decide rules/really treat as token")),
        (r"", None),
        (r"    %cell", (4, 9)),
        (r"# no %cell", None),
        (r"# %cell", None)
    ]
)
def test_get_cell_inj_span(test_line, expected):
    assert get_cell_inj_span(test_line) == expected


# TODO could clean up with textwrap dedent etc
@pytest.mark.parametrize(
    "test_source, expected, exception", [
        (r"""\
%cell
""", True, None),

        (r"""\
%cell
%cell
""", True, None),

        (r"""\
%cell;x
""", True, None),

        pytest.param(r"""\
%celll
""", None, None, marks=pytest.mark.xfail(reason="need to decide rules/really treat as token")),

        (r"""\
""", None, None),

        (r"""\
%cell l
""", True, None),

        (r"""\
x = 1
%cell
""", True, None),

        (r"""\
x = 1
if x==1:
    %cell
""", True, None),

        (r"""\
# no %cell
""", False, None),

        (r"""\
x = 1
if x==1:
    pass
    # no %cell
""", False, None),

        (r"""\
%cell # no %cell
""", True, None),

        (r"""\
%cell
# no %cell
""", NotImplemented, (ValueError, "mutually exclusive")),
    ])
def test_cell_injected_into_test(test_source, expected, exception):
    if exception is None:
        assert cell_injected_into_test(test_source) is expected
    else:
        with pytest.raises(exception[0], match=exception[1]):
            cell_injected_into_test(test_source)



from setuptools import setup, find_packages
from codecs import open
from os import path

from jupyter_packaging import (
    create_cmdclass, install_npm, ensure_targets,
    combine_commands, get_version
)

pjoin = path.join

name = 'nbcelltests'
here = path.abspath(path.dirname(__file__))
jshere = path.join(here, 'js')
version = get_version(pjoin(here, name, '_version.py'))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

requires = [
    'jupyterlab>=1.0.0',
    'nbval>=0.9.1',
    'nbconvert',
    'pytest>=4.4.0',
    'pytest-cov',
    'pytest-html>=1.20.0',
    'flake8',
    'parameterized',
]

dev_requires = requires + [
    'bump2version',
    'mock',
    'autopep8',
    'pytest-xdist ; sys_platform != "win32"',
    'beautifulsoup4'
]

data_spec = [
    # Lab extension installed by default:
    ('share/jupyter/lab/extensions',
     'js/lab-dist',
     'jupyterlab_celltests-*.tgz'),
    # Config to enable server extension by default:
    ('etc/jupyter',
     'jupyter-config',
     '**/*.json'),
]


cmdclass = create_cmdclass('js', data_files_spec=data_spec)
cmdclass['js'] = combine_commands(
    install_npm(jshere, build_cmd='build:all'),
    ensure_targets([
        pjoin(jshere, 'lib', 'index.js'),
        pjoin(jshere, 'style', 'index.css')
    ]),
)


setup(
    name=name,
    version=version,
    description='Cell-by-cell tests for JupyterLab',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/jpmorganchase/nbcelltests',
    author='The nbcelltests authors',
    license='Apache 2.0',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Framework :: Jupyter',
    ],

    cmdclass=cmdclass,

    keywords='jupyter jupyterlab',
    packages=find_packages(),
    install_requires=requires,
    extras_require={
        'dev': dev_requires
    },
    include_package_data=True,
    zip_safe=False,

)
